--tanh
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;
library ieee_proposed;
use ieee_proposed.fixed_float_types.all;
use ieee_proposed.float_pkg.all;

entity tanh is
  generic(
    we:positive:=4;
    wf: positive:=7
    );
  port(
    x: in std_logic_vector(we+wf downto 0);
    n: in std_logic_vector(3 downto 0);
    clk: in std_logic;
    start: in std_logic;
    sclr: in std_logic;
    z: out std_logic_vector(we+wf downto 0);
    rdy: out std_logic;
    underflow: out std_logic;
    overflow: out std_logic;
    invalid_op: out std_logic;
    divide_by_zero: out std_logic
    );
  end tanh;
  
  architecture behavioral of tanh is
  
    component factn
      generic(we: positive;
          wf: positive);
      port ( n : in integer range 0 to 15;
         data : out float(we downto -wf) );
    end component;
   
    constant unofp: float(we downto -wf):=to_float(1.0,we, wf);
    constant cerofp: float(we downto -wf):=to_float(0.0,we, wf);
    
    type estados is (t0,t1,t2,t3,t4,t5,t6,t7,t8);
    
    signal resultado: float(we downto -wf);
    signal estado, nuevoestado:estados:=t0;
    signal potx,potx_anterior,factorialn,div,par,impar: float(we downto -wf):=cerofp;
    signal n_asc: integer:=0;
    
    begin
      factn_0: factn
      generic map(
        we=>we,
        wf=>wf
      )
      port map(
        n=>n_asc,
        data=>factorialn
        );
        
      control: process(estado, start, sclr)
      begin
        if(sclr='1')then
          nuevoestado<=t0;
        else
        case estado is
          when t0=>
            if(start = '1') then
            nuevoestado<=t1;
            else
              nuevoestado<=t0;
            end if;      
          when t1=> 
            nuevoestado<=t2;           
          when t2=>nuevoestado<=t3;
          when t3=>nuevoestado<=t4;
          when t4=> nuevoestado <=t5;    
          when t5=> 
            if n_asc+1<conv_integer(n) then
              nuevoestado<=t1;            
            else
              nuevoestado<=t6;
            end if;
          when t6=> nuevoestado<=t7;
          when t7=> nuevoestado<=t8;
          when t8=> nuevoestado<=t0;
          
        end case;
        end if;
      end process;
      
      transiciones: process(clk)
      begin
        if rising_edge(clk) then
          estado<=nuevoestado;
        end if;
      end process;
      
      resultados_parciales: process(clk)
      variable z_fp_type : valid_fpstate;
      begin
        if rising_edge(clk)then
          case estado is
          when t0=>
           if sclr='1' then --clear
            potx<=cerofp;
            n_asc<=0;
            div<=cerofp;
            par<=cerofp;
            impar<=cerofp;
            z<=to_slv(cerofp);
            underflow<='0';
            overflow<='0';
            invalid_op<='0';
            divide_by_zero<='0';
            rdy<='0';
           elsif start='1' then --start
             n_asc<=0;
             div<=cerofp;
             par<=cerofp;
             impar<=cerofp;
             z<=to_slv(cerofp);
             underflow<='0';
             overflow<='0';
             invalid_op<='0';
             divide_by_zero<='0';
             rdy<='0';
           end if;
         when t1=>
            if n_asc=0 then --n=0
              potx<=unofp;
              potx_anterior<=unofp;
              elsif n_asc=1 then --n=1
                potx<=to_float(x,we,wf);
              else --n>1
                
                potx<=multiply(l=>to_float(x,we,wf),r=>potx_anterior,
                        round_style=>round_zero,guard=>0,
                        check_error=>true,denormalize=>true);
             -- end if;
           end if;
         when t2=>
         when t3=>
           potx_anterior<=potx;
           
           if n_asc>0 then
             if n_asc=1 then --n=1
               div<=to_float(x,we,wf);
             else --n>1
              div<=divide(l=>potx,r=>factorialn,round_style=>round_zero,
                      guard=>0,check_error=>true,denormalize=>true);
             end if;
            else --n=0
              div<=unofp;
           end if;
         when t4=>
         when t5=>
           if (n_asc mod 2)=0 then --n par
             par<=add(l=>par,r=>div,round_style=>round_zero,
                      guard=>0,check_error=>true,denormalize=>true);
            else --n impar
              impar<=add(l=>impar,r=>div,round_style=>round_zero,
                      guard=>0,check_error=>true,denormalize=>true);
            end if;
            n_asc<=n_asc+1;
          when t6=>
            resultado<=(divide(l=>impar,r=>par,round_style=>round_zero,
                      guard=>0,check_error=>true,denormalize=>true));
          when t7=>
          when t8=>
            z_fp_type := classfp (resultado, true);
            if z_fp_type=nan or z_fp_type=quiet_nan then
              divide_by_zero<='1';
            elsif z_fp_type=neg_inf or z_fp_type=pos_inf then
              overflow<='1';
            elsif z_fp_type=neg_denormal or z_fp_type=pos_denormal then
              underflow<='1';
            elsif z_fp_type=isx then
              invalid_op<='1';
            end if;
            z<=to_slv(resultado);
            rdy<='1';
          end case;
        end if;--start='1'
      end process;
    end behavioral;